<?php


// $Id: administer_predicates.inc,v 1.25.2.4 2006/07/13 13:06:44 dman Exp $

/**
 * @file
 *
 * Admin functions to administer the predicate terms definitions.
 *
 * Functions here relate only to the admin screens where
 * we can add and edit term definitions. None of the functions
 * in this file should be required to be loaded in a read-only
 * environment.
 *
 * The structure of this code was originally modelled on taxonomy.module,
 * and my RDF 'terms' are similar to Taxonomy 'terms',
 * I've discarded vocabulary support as being unneccessary, grouping is
 * handled by the ontologies.
 *
 * Terms can be defined and added to the term table,  preferably
 * via relationship_update_term();
 *
 * Pluggable Predicate Ontologies
 *
 * A module such as ontology_family may be created and enabled. It would
 * declare a bunch of predicates 
 * (IsFatherOf,  IsMotherOf,  IsParentOf, IsChildOf)
 * and their characteristics ( 
 *  IsMotherOf isSubPropertyOf IsParentOf,  
 *  IsChildOf isInverseOf IsParentOf
 *)
 * These predicates are added to the term list when the module is enabled,
 * and become available (on node edit) to anyone who wishes to use them. 
 * @see relationship_ontology_setup()
 *
 * To restrict or attach certain ontologies to only certain TYPES of node
 * (for example, you may create a custom node type for 'person', and only 
 * persons have family relations) the 'domain' of a term can be set.
 * This is analogous to associating node types with a vocabulary.
 *
 * Characteristics
 *
 * ... are axiomatic (basic,assumed) properties of predicates.
 *
 * Available characteristics should be used to relate one term with another.
 * Characteristics are modelled on the OWL definitions,  and include some
 * basic 'meanings' of term inter-relations. @see apply_characteristics.inc
 *
 * But sometimes that's not enough. 
 *
 * Predicate Ontology Modules can also define predicate logic lookups,
 * such as a routine that is called to calculate the results of a request to
 * 'IsCousinOf'.
 * This would return a meaningful result when the the system is asked for all
 * the cousins. This sort of lookup requres some coded,  repetative graph 
 * walking. If you wanted to encode the logic of a lookup HeirToTheThrone - 
 * this is where it would be done.
 *
 * These graph walks may be arbitrarily complex, and may need to call in 
 * many other inferred statements to get to their result. This on-demand
 * sort of query is different from the 'figure out what it means when we get it'
 * or on-save inferencing.
 *
 * The ontology libraries are also expected to provide appropriate ways of
 * displaying the relationships at node view time,  whether this be by publishing
 * a 'see also' sort of navigation block,  @see ontology_navigation.module 
 * or using hook_nodeapi() to modify the body output. 
 * @see ontology_revision_control.module
 * to see how isReplacedBy causes a warning to appear at the top of a page 
 * that is outdated.
 *
 * See the ontology modules for examples on how to make up
 * your own predicate programatically.
 *
 * 
 * Most of the time,  however the code treats 'terms' as objects,  
 * with a few properties and an array of characteristics.
 * Only when loading and serializing does this abstraction become important.
 *
 * Modules can also define their own characteristics and the
 * logical interpretation of what they imply,  however that 
 * is tricky and should be rare.
 *
 * @link predicates See Predicates for more about how I deal with predicate terms. @endlink
 * @see apply_characteristics.inc for how to code a characteristic.
 
 * @todo Use namespaces internally for indexing predicate terms? 
 * @todo be more selective
 *       how content types can choose various vocabularies to participate in.
 * ... partially done, using domain.
 *
 * @see ontologies
 * 
 * @ingroup docs
 * @package coders
 * @author Dan Morrison http://coders.co.nz/
 */


function relationship_admin(){
  $output = relationship_overview();
  print theme('page', $output);
}

/**
 * Menu callback; Edits or adds a predicate term
 */
function relationship_admin_term($op = 'add', $term = ''){
  # drupal_set_message("$op '$term'");

  if(!$term){$term = $_REQUEST['edit']['subject'];}
  if ($term){
    drupal_set_title("edit term $term");
    $o_term = relationship_get_term($term, TRUE);

    if ($op == t('Delete')){
      drupal_set_message("Deleted '$term'");
      drupal_goto('admin/relationship');
    }

  }
  else{
    drupal_set_title('add term');
    $o_term = (object) array (
      'type' => CLASS_PROPERTY,
    );
  }
  $term_form = _relationship_form_term($o_term);
  $output = drupal_get_form('relationship_predicate_form', $term_form);

  print theme('page', $output);
}


/**
 * Accept the form submission for a term and save the result.
 */
function relationship_predicate_form_submit($form_id, $form_values){
  relationship_save_term($form_values);
  relationship_flush_terms();
  return 'admin/relationship';
}

/**
 * @}
 * End of normal hooks, continue with some module-specific functions.
 *
 * Routines for displaying the admin forms.
 */

/**
 * @ingroup relationship The Relationship Management Interface under Drupal
 * @{
 */

/**
 * List a summary of all terms and the actions we can do on them.
 * @returns html table
 */
function relationship_overview(){
  $op = $_REQUEST['op'];
  $edit = $_POST['edit'];

  if ($op == t('Change Preferences')){
    foreach ($edit as $param => $val){
      variable_set($param, $val);
    }
  }

  if ($op == t('Hide Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms', FALSE);
  }
  if ($op == t('Show Advanced Ontologies')){
    variable_set('relationship_show_advanced_terms', TRUE);
  }

  $header = array (
    t('Name'), 
    t('Label'), 
    t('Characteristics'), 
    array (
    'data' => t('Operations'), 'colspan' => '3')
    );

  // Start with all known predicates, so we know which are left over as unsorted
  $terms = relationship_get_terms();
# debug_pre(array_keys($terms),0);

  // We'll list anything that's in an ontology
  $show_advanced = variable_get('relationship_show_advanced_terms', FALSE);
  // but filter some ontologies off the UI for convenience
  $hidden_ontologies = variable_get('relationship_hide_ontologies',array());
  

  // First enumerate the available, defined ontologies
  $ontologies = relationship_get_resources_matching_query(PREDICATE_TYPE, CLASS_ONTOLOGY);
  debug("Enumerating known ontologies :".join(", ",array_keys($ontologies)),1);
  foreach ($ontologies as $ontology_name => $ontology_def){
    // Skip the confusing technical ones unless admin has enabled that
    if ((!$show_advanced) && ( $hidden_ontologies[$ontology_name])){
      debug("Not listing members of $ontology_name, it's hidden by request ".join(', ',array_values($hidden_ontologies)) ,2); 
      $skip = TRUE;
      // even when skipping we need to iterate to cross these off the list.
      $hidden_ontologies[$ontology_name] = relationship_get_display_label($ontology_def);
    } 
    else {
      $skip = FALSE;
      unset($hidden_ontologies[$ontology_name]); // just for formatting the note below
      
      // prepare to display the block
      $data = relationship_get_display_label($ontology_def);
      $rows[] = array (
        array ( 'data' => '&nbsp;', 'colspan' => '5' )
      );
      $rows[] = array (
        array ( 'data' => $data, 'colspan' => '5',  'style' => 'font-weight:bold' )
      );
    }
      
   
    // Now retrieve all terms in this ontology, and format them
    debug("Listing terms defined in ontology '$ontology_name' " . ($skip ? "(hidden)" : ""), 2);

    // A loaded ontology should already know what its parts are.
    // however, this lookup also does the trick
    //  $terms_in_ontology = relationship_get_resources_matching_query(PREDICATE_IS_PART_OF, $ontology_def->uri,TRUE);

    $terms_in_ontology = array();
    if($ontology_def->hasPart){
      // always expect a list here
      if(! is_array($ontology_def->hasPart)){$ontology_def->hasPart = array($ontology_def->hasPart);}
      foreach($ontology_def->hasPart as $part_name ){
        // try the term cache (for efficiency) first
        if($term = relationship_get_term($part_name)){
           $terms_in_ontology[$part_name] = $term;
        }
      }
    }
dsm($terms_in_ontology);    
    foreach ($terms_in_ontology as $term){
      unset ($terms[$term->name]); // Tick this off as done
      
      if($skip){continue;}
      if( ($term->type != CLASS_PROPERTY) && (!$show_advanced) ){continue;}

      debug("Listing Term '$term->name' under $ontology_name", 3);
      $rows[] = _format_term_row($term);
    }
  }

  // Possible (but rare) that one of the original terms was not sorted into an ontology. 
  // List it anyway, so we can fix that. It'd be hard to find otherwise
  if ($terms){
    $rows[] = array (
      array ( 'data' => '&nbsp;', 'colspan' => '5' )
    );
    $rows[] = array (
      array ( 'data' => 'Unsorted', 'colspan' => '5', 'style' => 'font-weight:bold' )
    );
    foreach ($terms as $term){
      if (!$term){ continue; }
      $rows[] = _format_term_row($term);
    }
  }

  if (!$rows){
    $rows[] = array ( array ( 'data' => t('No terms available.'  ), 'colspan' => '5' ));
  }

  $output = theme('table', $header, $rows);

  if (variable_get('relationship_show_advanced_terms', FALSE)){
    $prefs .= t("
  				      <p>
  				        The terms and classes shown here are all resources that are either of 
  				        <strong>type owl:Property</strong>  or anything that 
  				        <strong>isPartOf</strong> any groupings of <strong>type owl:Ontology</strong>.
  				      </p><p>
                  The grouping displayed here is a result of a terms <em>membership</em>
                  in local <em>ontologies</em>. If 'domain' 'isPartOf' 'rdfOntology'
                  it will be grouped there. Terms can be members of multiple ontologies.
                  the <em>visibility</em> of ontologies can be adjusted in the settings.
  				        <br/>
                  Ontologies are useful to define a whole family of related terms,
                  and that ontology can be associated with a content type to
                  provide all the terms at once.
  				      </p>
  				      ");
    $button_action = t('Hide Advanced Ontologies');
  }
  else{
    $prefs .= t("
  				      <p>
  				        Some technical terms (%hidden_list) and the Ontologies and Classes
  				        may be currently <i>hidden</i> on this screen. 
  				      </p>
    				    ",
      array('%hidden_list' => join(', ',$hidden_ontologies) )
    );
    $button_action = t('Show Advanced Ontologies');
  }
  $buttons = array(
    l($button_action, 'admin/relationship', array (), 'op=' . $button_action),
    l('More Settings', 'admin/settings/relationship')
  );
  $prefs .= theme('item_list',$buttons);

  $output .= $prefs;

  $output .= t("
        		    <p>
        		      The best way to define new terms is to install an ontology module which
        		      defines its own predicates, or install the <b>rdf_ontology.module</b> and
        		      import globally-defined terms from the web using that.
        		      <br/>Plugin Ontology modules should be available from the Drupal module repository.
        		    </p>
            		");

  return $output;
}

/**
 * Format display of the defined term or characteristic properties.
 * Returns an array for use as a tr.
 * helper function for relationship_overview()
 */
function _format_term_row($term){
  $relationship = "";

  /*
  static $all_characteristics;
  if (!$all_characteristics)
  {
    $all_characteristics = relationship_get_all_characteristics();
  }

  foreach ($all_characteristics as $name => $characteristic)
  {
    // Only list INTERESTING characteristics here. 
    // Interesting means the weight of the predicate is > 4;

    if ($term->$name && (0 + $characteristic->weight) > 4)
    {
      $relationship .= '<small class="term-characteristic"  >' . $name . ($term-> $name ? ":" . '</small><small class="term-characteristic-value">' . $term->name : '') . '</small><br/> ';
    }
  }
  */
  if(! relationship_term_is_valid($term)){
    drupal_set_message("Problem with term definition ".print_r($term,1),'error'); 
  }
  
  if(($term->type != CLASS_PROPERTY)){
    $relationship = $term->type;
  }

  $actions = l(t('edit term'), "admin/relationship/term/edit/", NULL, "edit[subject]=" . urlencode($term->uri));
  $actions .= " " . l(t('delete'), "admin/relationship/term/delete/" . $term->name);

  return array (
    ($term->uri ? "<a href='{$term->uri}' title='Formal definition of this term'>{$term->name}</a>" : $term->name),
    l($term->label, "relationship/view/{$term->name}", array ('title' => $term->description) ),
    //    '<acronym title="'.check_plain((string)$term->description).'" >'.check_plain((string)$term->label).'</acronym>',
    $relationship, 
    $actions
  );
}

/**
 * Display an edit form for a predicate 'term'.
 * 
 * Set its name,  label,  characteristics and descriptions here.
 * Should also work for 'new' terms.
 * @param $edit assoc array of edit fields
 * @return HTML form
 */
function _relationship_form_term($term_def = array ()) {
  $edit = (array) $term_def;
  $assist_scripts = "\n/* Little form usability additions */\n";

  $form['label'] = array (
    '#type' => 'textfield',
    '#title' => $term_def->type .' '. t('label'),
    '#default_value' => $edit['label'],
    '#size' => 50,
    '#maxlength' => 64,
    '#description' => t("
            		          The human-readable predicate for this resource. 
            		          Use a descriptive phrase with leading caps. 
            		          Choose one from <a href='http://dublincore.org/documents/dcmi-terms/' 
            		          target='_new'>the Dublin Core list</a> if possible. 
            		          For predicates, this should be a descriptive sentence fragment that will make grammatical
            		          sense of the statement 'Document A [....] Document B'.  
            		          Example: 'Is Part Of', 'References', 'Is Replaced By'.
            		        "), 
    '#attributes' => NULL,
    '#required' => TRUE,
    );

  // NOTE, although the statement us about the identifier, the object has the property 'name'
  // they mean the same thing.
  $form['name'] = array (
    '#type' => 'textfield',
    '#title' => $term_def->type .' '.t('identifier'),
    '#default_value' => $edit['name'], 
    '#size' => 50, 
    '#maxlength' => 64,
    '#description' => t("
            		          The internal name/id. 
            		          May be the same as,  or a shorthand (CamelCase) form of the label. 
            		          Should match the <a href='http://dublincore.org/documents/dcmi-terms/' 
            		          target='_new'>Dublin Core naming convention</a> if possible. 
            		          Example: 'isPartOf', 'References', 'IsReplacedBy'.
            		        "), 
    '#attributes' => array ( 'onclick' => "update_name(this)" ), 
    '#required' => TRUE,
  );
  $assist_scripts .= "function update_name(element){if(!element.value){element.value = element.form.elements['edit-label'].value.replace(/ /g,'');}}\n";

  $form['description'] = array (
    '#type' => 'textarea',
    '#title' => t('Description'),
    '#default_value' => $edit['description'],
    '#cols' => 60,
    '#rows' => 5,
    '#description' => t("
            		          A description of what relationship the term is supposed to describe.
            		        "),
    '#attributes' => array ('onclick' => "update_description(this)"),
  );
  $assist_scripts .= "function update_description(element){if(!element.value){element.value = 'This resource '+element.form.elements['edit-label'].value+' the referenced resource.'}}\n";

  if ($term_def->derived)
  {
    $form[] = array (
      '#value' => t("
          This property is <em>derived</em> which means it is never
          stated explicitly by a user, but is set only through deduction
          based on other relationships. Exactly what those relationships 
          <em>are</em> is defined in the characteristics below.
        ")
    );
  }
  $form['uri'] = array (
    '#type' => 'textfield',
    '#title' => t('URI' ),
    '#default_value' => $edit['uri'], 
    '#size' => 50,
    '#maxlength' => 64, 
    '#description' => t("
            		          A reference to the definative resource (schema) that describes this term.
            		          A technical definition of what properties this metadata term implies. 
            		          Eg: 
            		            'http://purl.org/dc/terms/replaces', 
            		            'http://www.w3.org/TR/html4/types.html#type-links' . 
            		          Should probably be some implimentation of 
            		          <a href='http://www.w3.org/TR/owl-features/' title='Web Ontology Language'>OWL</a>
            		          which can then define term properties (characteristics) like 
            		          'commutative' and 'associative'. 
            		          Or you can just make your own URI up or leave it blank.
            		          Even if its not a formal schema, it is still used as a namespace.
            		          Schemas can be found at 
            		          <a href='http://www.schemaweb.info' target='_new'>Schemaweb</a>
            		          and other places.
            		        "),
    '#attributes' => array ('onclick' => "update_uri(this)"),
    '#required' => FALSE,
  );
  $assist_scripts .= "function update_uri(element){if(!element.value){element.value = element.form.elements['edit-" . 'name' . "'].value;}}\n";

  $type_urls = relationship_query(NULL, PREDICATE_TYPE, CLASS_CLASS, 'subject');
#debug("all things of type class");
#debug_pre($type_urls);
  foreach ($type_urls as $url){
    $types[$url] = relationship_get_shortname($url);
  }
  // use longnames under the hood, display shortnames.
  $form['type'] = array (
    '#type' => 'select',
    '#title' => t('Type'), 
    '#default_value' => relationship_get_longname($edit['type']), 
    '#options' => $types, 
    '#description' => t("
            		          The type of resource this is. 
            		          Predicate Terms are of type <strong>Property</strong>
            		          but advanced options are possible if OWL classes are being defined
            		        "),
  );

  $form['weight'] = array (
    '#type' => 'weight',
    '#title' => t('Weight' ), 
    '#default_value' => $edit['weight'],
    '#delta' => 10,
    '#description' => t(" Importance for display and listings. "),
    );
  $form['help'] = array (
    '#value' => t(" 
                  It's not always appropriate for all terms to show up on all screens,
                  What gets displayed where can be controlled with a
                  combination of accurate 'domain' settings, and the
                  use of ontology groupings.

                  Term 'nav:last' would be an <em>implied</em> property only (derived from
                  nav:next) and although it provides part of an ontology, it is not 
                  something the user should override. 
              "),
  );

  // Use the node relationship form here to apply further statements to this form
  // Prepare the appropriate drop-downs 

  $characteristics = relationship_get_all_characteristics($edit['type']);
  # debug_pre(array_keys($characteristics));

  $predicates = array_merge(
    array ( '' => '' ),
    relationship_get_terms()
  );

  // flatten this list - it's to be used as an option drop-down
  foreach ($predicates as $i => $p){ $predicate_list[$i] = $p->name; }

  $form[] = relationship_form($term_def, $characteristics, $predicate_list);

  $form['Submit'] = array (
    '#type' => 'submit',
    '#value' => t('Submit'),
  );

  if ($edit['name']){
    $form['Delete'] = array (
      '#type' => 'submit',
      '#value' => t('Delete'),
    );
  }
  else{
    $form['destination'] = array (
      '#type' => 'hidden',
      '#value' => $_GET['q'],
    );
  }

  //
  // Format the help about characteristics
  //
  $glossary = "<p>
      		      There may be relationship <em>between</em> these relationship,  eg 
      		      '<strong>References</strong>' is the <em>inverse</em> of 
      		      '<strong>Is Referenced By</strong>',  and a '<strong>Is Part Of</strong>'
      		      property is <em>transitive</em> because it's implied that a part of a 
      		      part also a part of the whole.
      		  </p><p>
      		      For more detail,  refer to the 
      		      <a href='http://www.w3.org/TR/2004/REC-owl-guide-20040210/#PropertyCharacteristics'>
      		      Web Ontology Language</a>. 
      		      If used,  this extra information can be used as rules for 'graph walking'
      		      to deduce implicit metadata for items without stating them explicitly 
      		      each time,  eg a page can infer which pages reference it without 
      		      maintaining a list of referees itself.
      		  </p>\n";

  foreach ($characteristics as $char_name => $char_def){
    $glossary .= "\n<dt><a href='{$char_def->uri}' title='See the Formal definition of this characteristic'>{$char_def->name} : {$char_def->label}</a></dt>\n<dd>{$char_def->description}</dd>\n";
  }

  $glossary .= '</dl>';

  $form[] = array (
    '#type' => 'fieldset',
    '#title' => t('About Property Characteristics' ),
    '#description' => t($glossary),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    );

  /* Advanced diagnostic dump */
  if ($edit && variable_get('relationship_show_advanced_terms', FALSE)){
    $dump = var_export($edit, 1);
    $form[] = array (
      '#value' => "<h4>'Advanced Terms' is enabled in the settings</h4>Dumping the structure for inspection. Turn off 'advanced' to stop this" 
    );
    $form[] = array (
      '#value' => "<div><textarea rows='" .
       (count(explode("\n",
      $dump
    )) + 2) . "' cols='80'>" . $dump . "</textarea></div>");
  }

  if ($edit['uri'] && function_exists('relationship_rdf_render_rdf_about')){ 
  // RDF interface is optional
    $form['RDF'] = array ( 
      '#type' => 'fieldset',
      '#title' => 'RDF',
      '#collapsible' => TRUE, 
      '#collapsed' => TRUE, 
    );
    $form['RDF']['title'] = array(
      '#value' => l('RDF', "relationship/rdf/" , 
        array ('title' => t('The pure RDF-metadata about this resource.') ),
        "edit[uri]=".urlencode( $edit['uri'] )
      )
    ); 
    $form['RDF']['dump'] = array( '#value' =>              
      '<div style="width:500px; overflow:scroll">'
      . relationship_rdf_render_rdf_about($edit['uri'])
      . '</div>'
    );
  }

  drupal_set_html_head("<script type='text/javascript' >\n//<![CDATA[\n$assist_scripts\n//]]>\n</script>");

  return $form;
}

/**
 * Save the input from an edit term form.
 *
 * Deal with a few special cases,  then serialize the data
 * @param $edit assoc array of edit fields
 * @returns assoc array,  possibly slightly modified for the serialization
 */
function relationship_save_term($edit)
{
  if ($_REQUEST['op'] == t('Delete')){
    drupal_set_message(t('Deleting relationship term "%name" ', array (
      '%name' => $edit['name']
    )));
    return relationship_delete_term($edit['name']);
  }

  $term = (object) $edit;

  // among other things, this reads the flat form values back into statement structure
  $term = relationship_resource_validate($term);

  $term->type = relationship_get_shortname($term->type);

  $intrinsic_statements =  relationship_array2statements($edit, $edit['uri'], NULL, TRUE);
  $term->statements = array_merge($intrinsic_statements, $term->statements);

  // Special-case the 'Inverse' property
  // tell the other relationship to point back to us
  if ($other_predicate_assertion = $term->statements['inverseOf']){
    $other_predicate = $other_predicate_assertion['object'];
    drupal_set_message("As it's an inverse,  I'm linking target term $other_predicate back to this one as $characteristic also.");
    $other_term = relationship_get_term($other_predicate);
    $other_term->characteristics['inverseOf'] = $term->name;
    $other_term->inverseOf = $term->name;

    relationship_save_resource($other_term);
  }

  if ($term->name){
    relationship_replace_resource_statements($term);
    drupal_set_message( t("The term %term has been updated.", 
      array ('%term' => theme('placeholder', $term->name ) )
    ));
  }
  else{
    drupal_set_message( t("Unnamed term was not updated") );
    debug_pre($term);
  }
  // remember to relationship_flush_terms(); if needed

  return $term;
}


function relationship_delete_term($name, $extreme_prejudice=FALSE) {
  if(!$name){debug('Invalid argument ('.$name.') to relationship_delete_term()',1);return;}
  debug("Deleting the term '$name'",2);

  $uri = relationship_get_longname($name);
  relationship_resource_delete_statements($uri);
  
  if($extreme_prejudice){
    $ds = _relationship_datastore();
    drupal_set_message("Deleting all statements using the term '$name'",2);
    $ds->delete(NULL,$name,NULL) ;
    $ds->delete(NULL,$uri,NULL) ;
  }
  // remember to flush after yourself, this term may have left smears in the cache
}


/**
 * @}
 */



/**
 * Performs a consistancy check on the database.
 *
 * Removes references to non-existant nodes
 * Allows a refresh of the predicate data.
 * Menu callback
 */
function relationship_check_database(){
  $op = $_REQUEST['op'];
  $edit = $_REQUEST['edit'];

  /* Clean up nids that don't exist */

  if ($op == t('Drop')){
    drupal_set_message('Dropping ALL data from database');
    db_query('DELETE FROM {relationship_statement}');
    relationship_flush_terms();
  }

  $out = "<p>
          <h2>Cleanup database</h2>
          Although it should not happen, it's possible that we could be left
          with references concerning nodes that no longer exist. This page can 
          check and clean these up. It's really just an admin function for
          development.
          </p>
  		  ";

  // statement triples cannot be empty or null
  $invalid_statement_query = " (TRIM(subject) = '') OR (TRIM(predicate) = '') OR (TRIM(object) = '') ";
  $query = "SELECT s.%s as nid FROM {relationship_statement} s LEFT JOIN {node} n ON ( s.%s = n.nid) WHERE s.%s > 0 AND n.nid IS NULL GROUP BY s.%s ";
  $misdirected_subjects = db_query($query, 'subject', 'subject', 'subject', 'subject');
  $out .= '<p>Found ' . db_num_rows($misdirected_subjects) . " statements in DB with an invalid subject<br/>";

  $form = array ();
  while ($s = db_fetch_object($misdirected_subjects)){
    $out .= 'node:' . $s->nid . '<br/>';
    $form['nid'][''] = array (
      '#type' => 'hidden',
      '#value' => $s->subject,
    );
  }
  $out .= '</p>';

  $misdirected_objects = db_query($query, 'object', 'object', 'object', 'object');
  $out .= '<p>Found ' . db_num_rows($misdirected_objects) . " statements in DB with an invalid object<br/>";

  $query = "SELECT rsid FROM {relationship_statement} WHERE $invalid_statement_query";
  $invalid_statements = db_query($query);
  $out .= '<p>Found ' . db_num_rows($invalid_statements) . " statements in DB that have invalid options<br/>";

  while ($s = db_fetch_object($misdirected_objects)){
    $out .= 'node:' . $s->nid . '<br/>';
    $form['nid'][''] = array (
      '#type' => 'hidden',
      '#value' => $s->nid,
    );
  }
  $out .= '</p>';
  $form['nid'][''] = array (
    '#type' => 'submit',
    '#value' => t('Clean Up'),
  );

  $out .= drupal_get_form('relationship_database_maintainence', $form);

  /* Check remote URLs? */

  $out .= ' ' . l(t('Drop'), 'admin/relationship/tidy', array (), 'op=' . t('Drop'));

  $out .= "<p>
  		      <h2>Reset Axiomatic Predicates</h2>
  		      In the database, there are a set of underlying 'axioms' defining
  		      The very definition of the terms used as predicates. 
  		      These sometimes need tweaking. Use the below import buttons to
            make sure they are up-to-date.
     		  </p>";



  // Importing from Ontologies
  $import_command = "Import";
  $schema_files = file_scan_directory( dirname(drupal_get_filename('module','relationship')),"^[^\.].*\.rdf",array('.', '..', 'CVS'),NULL,TRUE );
  foreach($schema_files as $i => $schema_file){
    $imports[$i] = "Import ".l($schema_file->name, 'admin/relationship/tidy', NULL, "op=$import_command&edit[source]=".$schema_file->filename );
  }
  $out .= theme_item_list($imports,"Import Additional Ontologies");

  if($op == $import_command){
    $source = $edit['source'];
    require_once('relationship_rdf.module'); // use this for the import, even if it's not enabled yet
    relationship_rdf_read_from_source($source);
    relationship_flush_terms();
  }

  print (theme('page', $out));
}


/**
 * Impliment the requested database maintainence.
 * @see relationship_check_database()
 */
function relationship_database_maintainence_submit($form_id, $form_values){
  $nids = $form_values['nid'];
  if (is_array($nids)){
    drupal_set_message('Removing references to ' . count($nids) . ' objects..');
    foreach ($nids as $nid){
      relationship_resource_delete_statements($nid, TRUE);
    }
  }
  $invalid_statement_query = " (TRIM(subject) = '') OR (TRIM(predicate) = '') OR (TRIM(object) = '') ";
  $query = "DELETE FROM {relationship_statement} WHERE $invalid_statement_query";
  db_query($query);
  drupal_set_message(t('Cleaned up database'));
}


/**
 * @ingroup relationship_api
 * Public Routines you might want to use from outside.
 * Mostly for adding, retrieving and updating term entries.
 * @{
 */

/**
 * Check that the given object is suitable for use as a predicate term.
 *
 * Fill in any gaps, and abstract the current properties into statements.
 * relationship_save_term() is the UI form that usually calls this.
 * This is an actin that actually does stuff, distinct from
 * relationship_term_is_valid() which is just a check. 
 * @param $term object definition
 * @return success
 * @ingroup relationship_api
 * @see relationship_delete_term()
 * @see predicates The predicate term object definition.
 */
function relationship_validate_term(& $edit){
  // A term object is really an abstract collection of statements.
  // Saving the statements collectively means saving the term.

  if (!$edit['name']){ // This is an API, more paranoia
    drupal_set_message('Cannot work with a term with no name identifier at all! Problem updating term. ' . print_r($edit, 1), 'error');
    return;
  }

  debug('validating term ' . $edit['name'], 2);

  if (!$edit['uri'])
  { // This is an API, more paranoia
    $edit['uri'] = $edit['name'];
    // Must have a uri now, should create a localhost id I guess
    debug(t('
			        Setting term URI to just "%uri" - this means that the term is 
			        basically only valid in the context of the current site. 
			        This is OK for made-up relationships, but means this bit of data
			        is harder to share with other servers, as it has no namespace.
			    ', array ('%uri' => $edit['uri'] )
    ), 1);
  }

  $edit['type'] = ($edit['type']) ? $edit['type'] : CLASS_PROPERTY;
  // This statement is what makes a predicate a predicate

  // check if there's an older version
  $old_term = relationship_get_term($edit['name']);
  if ($old_term && ($edit['uri'] != $old_term->uri)){
    // If they've changed the URI, it's annoying, as that's the UID for the term
    // Need to clear out everything that it used to know.
    // The update below will replace it with new values. Hopefully it won't break anything.
    relationship_delete_term($old_term->uri);
  }

  #debug_pre($term);
  debug("Generating statements from term definition", 3);
  $edit['statements'] = relationship_array2statements($edit, $edit['uri'], relationship_get_property_map($edit['type']), NULL, 'From Term Definition Validation');

  // For convenience (but not efficiency) I'll try to use shortnames when saving statements
  // about my own ontologies. This will have side-effects later on.
  foreach ($edit['statements'] as $i => $s){
    relationship_validate_statement($edit['statements'][$i]);
  }

  debug("<h2>Validated term " . $edit['name'] . "</h2>", 1);
  #debug_pre($edit);

  return $edit;
}

/**
 * Returns just true or false if the predicate term has no troubles.
 */
function relationship_term_is_valid($term){
  if (!$term->name){
    $problem = 'Cannot work with a term with no name identifier at all.';
  }
  if (!$term->uri){
    $problem = 'Term has no URI.';
  }
  if (!$term->label){
    $problem = 'Term has no label.';
  }

  if($problem){
    drupal_set_message($problem . '<pre>'.print_r($term, 1).'</pre>', 'error');
    return FALSE;
  }
  return TRUE;
}

/**
 * Return a list describing all possible properties that may apply to a type.
 * 
 * * This list may be modified by plugins adding extra logic and making
 * appropriate term definitions claiming their new characteristic has 'domain'
 * of 'property' or 'ObjectProperty'.
 *
 * As this list is pretty stable, it's cached. The cache should be flushed
 * whenever a term is modified, and it'll be rebuilt if empty.
 *
 * @param $flush if set, dump the cache and force a re-read, re-calculation of the data
 * @return assoc array of available characteristic definitions. 
 *         Indexed by the characteristic shortname.
 *         The definition should include the label and the URI of the characteristic
 * @see apply_characteristics.inc
 * 
 */
function relationship_get_all_characteristics($type = CLASS_PROPERTY, $flush = FALSE){
// TODO deprecate this, its a waste of time?
// At least do some metrics on how often it's needed - is caching right?'
  $characteristic_cache_key = "relationship_characteristics";
  if($flush){
    cache_set($characteristic_cache_key,NULL);
  }
  $cached = cache_get($characteristic_cache_key);
  if ($cached->data){
    return unserialize($cached->data);
  }
  debug("Retrieving all properties that may apply to items of class $type", 2);

  // Characteristics are those predicates whose 'domain' 
  // (things it applies to) is 'property'. 
  // $characteristic_list = relationship_query( NULL, PREDICATE_DOMAIN, CLASS_PROPERTY, 'subject' );
  //
  // equivalent to:
  $characteristic_list = relationship_query($type, PREDICATE_CAN_HAVE_PROPERTY, NULL, 'object');
  # debug_pre($characteristic_list, 0);

  // Thanks to inference, this list returns duplicates, due to the keying on full statement triples.
  // there may be more than one reason why the statement is returned, we've retrieved them ALL
  // Refactor this list to use all longnames
  $characteristics = array ();
  foreach ($characteristic_list as $k => $s){
    $s = relationship_get_longname($s);
    $characteristics[$s] = relationship_get_term($s);
  }
  cache_set($characteristic_cache_key, serialize($characteristics));
  return $characteristics;
}

/**
 * @}
 */

/**
 * Implimentation of (proposed) hook_setup() .
 *
 * Insert (or remove) a set of term definitions into the predicate table.
 * Called once on first-run,  or optionally with a 'refresh' flag to force a reversion.
 * 
 * This is a helper utility function for use by pluggable ontologies. If an
 * ontology declares an ontology_defined_terms() sort of function,  which returns 
 * an array of term defs,  AND this pluggable ontology has just been enabled,
 * insert those term definitions into the datastore for use.
 * Or,  if the ontology has just been disabled (through the use of modules admin)
 * remove the terms. This is non-destructive,  and leaves any defined relationships
 * behind,  although their logic will probably not work any more.
 *
 * Internally,  we set a variable flag {modulename}_db_initialized to monitor
 * the installation status of a module.
 *
 * @param $module_name
 * @param $refresh Force the installation, even if it's apparently already there.
 * @ingroup ontologies
 * @todo rework this with the hook.install method - still need an uninstall
 */
function relationship_ontology_setup($module_name, $refresh = 0){
  // I'm passing the module name rather than just the terms array itself
  // in case I want to do anything more advanced with this function some time.
drupal_set_message("TODO review the function relationship_ontology_setup()",1);
  $get_terms = $module_name . '_defined_terms';
  // if the module is turned on,  but not ready for action, 
  // make sure the db lists its published predicates.
  if (function_exists($get_terms)){
    if (module_exist($module_name) && ((!variable_get($module_name . '_db_initialized', FALSE)) || $refresh))
    {
      drupal_set_message(t('Setting up new module %module_name', array (
        '%module_name' => $module_name
      )));

      $new_terms = $get_terms ();
      relationship_import_definitions($new_terms);
      relationship_flush_terms();

      drupal_set_message("Initialized node relationship (predicate) terms used by $module_name. The predicate_list should now include " . implode(",  ", array_keys($get_terms ())));
      variable_set($module_name . '_db_initialized', TRUE);
    }
    else {
      if (variable_get($module_name . '_db_initialized', FALSE) && (!module_exist($module_name)))
      {
        // We've just been turned off. Clean up a little. Remove terms from database. 
        // Leave relationships using that term however. 
        // This may orphan them but it's better than data loss.
        foreach ($get_terms () as $term_name => $term_def){
          relationship_delete_term($term_name);
        }
        relationship_get_terms('', TRUE); // this will flush the terms cache for us

        variable_del($module_name . '_db_initialized');
        drupal_set_message("
          Disabled the node relationship (predicate) terms used by 
          $module_name. The terms " .
          implode(",  ", array_keys($get_terms ())) . " 
          have been disabled,  however any statements still using those terms
          were left behind on the nodes. 
          To vaporize them totally,  you would have been best to DELETE the 
          term using the term administration interface - if that is your intent.
			  ");
      }
    }
  } // finished updating terms definitions.
}

/**
 * Import a structured array of info 
 * as a set of statements.
 *
 * When given an optional $import_template, certain terms will be mapped
 * from the structure array keys to statement predicates.
 * as in @see relationship_array2statements()
 *
 * This will also CLEAR OUT any previous statements made about
 * the subjects of the import.
 *
 * The new statements will be indexed on the uri or the identifier of the 
 * given set of data - or the $subject if given. In that order of precedence.
 * If not told otherwise, this assumes that the import statements define a TERM (type:CLASS_PROPERTY)
 *
 * @see relationship_import_definition() for the singular version of this action
 * @param $defs an array of definition arrays.
 */
function relationship_import_definitions($defs){
  if (!$defs){
    drupal_set_message(t('No term definitions to import, an empty array passed in when trying to import definitions from a module ontology.'), 'error');
    return;
  }

  foreach ($defs as $subject => $subject_data){
    if ($subject_data){
      if (!$subject_data['name']){
        $subject_data['name'] = $subject;
      }
      if ($term = relationship_validate_term($subject_data)){
        // validate has figured it out into statements
        relationship_replace_resource_statements($term);
        drupal_set_message(t('Imported a set of statements defining what "%term" means.', array (
          '%term' => $subject_data['name']
        )));
      }
      else{
        drupal_set_message(t('Trouble importing the definition of "%subject" - apparently the supplied data did not return as a valid term.', array (
          '%subject' => $subject
        )));
      }
    }
  }
  // Usually need to flush the predicate terms cache right about now, we have some new terms.
}